from unittest import TestCase

from pactman.mock.provider import Provider


class ProviderTestCase(TestCase):
    def test_init(self):
        result = Provider('TestProvider')
        self.assertIsInstance(result, Provider)
        self.assertEqual(result.name, 'TestProvider')



import json
import pathlib
from itertools import chain
from unittest.mock import Mock

import pytest
import requests
import semver
from restnavigator import Navigator

from pactman.verifier.broker_pact import BrokerPact, BrokerPacts, pact_id
from pactman.verifier.matching_rule import log, rule_matchers_v2, rule_matchers_v3
from pactman.verifier.result import Result
from pactman.verifier.verify import Interaction, RequestVerifier, ResponseVerifier


BASE_DIR = pathlib.Path(__file__).absolute().parents[0]


def all_testcases(path):
    for entry in path.iterdir():
        if entry.is_file() and entry.suffix == '.json':
            yield str(entry)
        elif entry.is_dir():
            yield from all_testcases(entry)


@pytest.fixture
def fake_interaction():
    return {
        'description': 'dummy',
        'request': {'method': 'GET', 'path': '/users-service/user/alex'},
        'response': {'headers': {}, 'status': 200}
    }


@pytest.fixture
def fake_pact(fake_interaction):
    return {
        'provider': {'name': 'SpamProvider'},
        'consumer': {'name': 'SpamConsumer'},
        'interactions': [fake_interaction],
        'metadata': {'pact-specification': {'version': '2.0.0'}},
    }


@pytest.fixture
def mock_pact():
    def create_mock(version):
        return Mock(provider='SpamProvider', consumer='SpamConsumer', version=semver.parse(version))
    return create_mock


@pytest.fixture
def mock_result():
    result = Result()
    result.fail = Mock(return_value=False)
    return result


def test_pact_id():
    assert pact_id(1) == repr(1)


def test_pact_loading(monkeypatch, fake_pact):
    p = BrokerPacts('SpamProvider', 'http://broker.example/')
    mock_pact = Mock(fetch=Mock(return_value=fake_pact))
    mock_provider = Mock(fetch=lambda: None, __getitem__=lambda s, k: [mock_pact])
    monkeypatch.setattr(Navigator, 'hal', lambda url, default_curie=None: {
        'latest-provider-pacts': lambda provider=None: mock_provider
    })

    for e in p.consumers():
        assert str(e) == '<Pact consumer=SpamConsumer provider=SpamProvider>'
        assert e.provider == 'SpamProvider'

    for e in p.all_interactions():
        assert e.description == 'dummy'


def test_pact_publish(fake_pact):
    mock_result = Mock()
    mock_pact = {'publish-verification-results': mock_result}
    p = BrokerPact(fake_pact, Mock(), mock_pact)
    p.publish_result('1.0')
    mock_result.create.assert_called_once()


def test_pact_file_loading(fake_interaction):
    p = BrokerPact.load_file(str(BASE_DIR / 'testcases-version-3' / 'dummypact.json'))
    assert str(p) == '<Pact consumer=SpamConsumer provider=SpamProvider>'
    assert p.provider == 'SpamProvider'
    for e in p.interactions:
        assert e.description == 'dummy'


def test_interaction(mock_pact, mock_result, fake_interaction):
    i = Interaction(mock_pact('2.0.0'), fake_interaction, mock_result)
    assert repr(i) == 'SpamConsumer:dummy'


def test_interaction_verify_get(monkeypatch, mock_pact, mock_result, fake_interaction):
    monkeypatch.setattr(requests, 'post', Mock())
    monkeypatch.setattr(requests, 'get', Mock())
    monkeypatch.setattr(ResponseVerifier, 'verify', Mock())
    i = Interaction(mock_pact('2.0.0'), fake_interaction, mock_result)
    i.verify('http://provider.example/', 'http://provider.example/pact-setup/')

    requests.post.assert_not_called()
    requests.get.assert_called_with('http://provider.example/users-service/user/alex', headers={})
    i.response.verify.assert_called()


def test_interaction_verify_method_not_supported(monkeypatch, mock_pact, mock_result, fake_interaction):
    monkeypatch.setattr(requests, 'get', Mock())
    monkeypatch.setattr(ResponseVerifier, 'verify', Mock())
    fake_interaction['request']['method'] = 'FLEBBLE'
    i = Interaction(mock_pact('2.0.0'), fake_interaction, mock_result)
    i.verify('http://provider.example/', 'http://provider.example/pact-setup/')
    i.result.fail.assert_called_with("Request method FLEBBLE not implemented in verifier")
    i.response.verify.assert_not_called()


def test_interaction_verify_qs(monkeypatch, mock_pact, mock_result, fake_interaction):
    monkeypatch.setattr(requests, 'post', Mock())
    monkeypatch.setattr(requests, 'get', Mock())
    monkeypatch.setattr(ResponseVerifier, 'verify', Mock())
    fake_interaction['request']['query'] = 'a=b&c=d'
    i = Interaction(mock_pact('2.0.0'), fake_interaction, mock_result)
    i.verify('http://provider.example/', 'http://provider.example/pact-setup/')

    requests.get.assert_called_with(
        'http://provider.example/users-service/user/alex',
        params=dict(a=['b'], c=['d']),
        headers={},
    )
    i.response.verify.assert_called()


def test_interaction_verify_post(monkeypatch, mock_pact, mock_result, fake_interaction):
    monkeypatch.setattr(requests, 'post', Mock())
    monkeypatch.setattr(requests, 'get', Mock())
    monkeypatch.setattr(ResponseVerifier, 'verify', Mock())
    fake_interaction['request']['method'] = 'POST'
    fake_interaction['request']['body'] = 'spam'
    i = Interaction(mock_pact('2.0.0'), fake_interaction, mock_result)
    i.verify('http://provider.example/', 'http://provider.example/pact-setup/')

    requests.get.assert_not_called()
    requests.post.assert_called_with(
        'http://provider.example/users-service/user/alex',
        json='spam',
        headers={},
    )
    i.response.verify.assert_called()


def test_interaction_verify_post_unsupported_content_type(monkeypatch, mock_pact, mock_result, fake_interaction):
    monkeypatch.setattr(requests, 'post', Mock())
    monkeypatch.setattr(ResponseVerifier, 'verify', Mock())
    fake_interaction['request']['method'] = 'POST'
    fake_interaction['request']['headers'] = {'Content-Type': 'spam/ham'}
    i = Interaction(mock_pact('2.0.0'), fake_interaction, mock_result)
    i.verify('http://provider.example/', 'http://provider.example/pact-setup/')
    i.result.fail.assert_called_with("POST content type spam/ham not implemented in verifier")

    requests.post.assert_not_called()


def test_interaction_verify_delete(monkeypatch, mock_pact, mock_result):
    monkeypatch.setattr(requests, 'delete', Mock())
    monkeypatch.setattr(ResponseVerifier, 'verify', Mock())
    fake_interaction = {
        'description': 'dummy',
        'request': {'method': 'DELETE', 'path': '/diary-notes/diary-note/1'},
        'response': {'headers': {}, 'status': 200}
    }
    i = Interaction(mock_pact('2.0.0'), fake_interaction, mock_result)
    i.verify('http://provider.example/', 'http://provider.example/pact-setup/')

    requests.delete.assert_called_with('http://provider.example/diary-notes/diary-note/1', headers={})
    i.response.verify.assert_called()


def test_interaction_verify_put(monkeypatch, mock_pact, mock_result, fake_interaction):
    monkeypatch.setattr(requests, 'put', Mock())
    monkeypatch.setattr(requests, 'get', Mock())
    monkeypatch.setattr(ResponseVerifier, 'verify', Mock())
    fake_interaction['request']['method'] = 'PUT'
    fake_interaction['request']['body'] = 'spam'
    i = Interaction(mock_pact('2.0.0'), fake_interaction, mock_result)
    i.verify('http://provider.example/', 'http://provider.example/pact-setup/')

    requests.get.assert_not_called()
    requests.put.assert_called_with(
        'http://provider.example/users-service/user/alex',
        json='spam',
        headers={},
    )
    i.response.verify.assert_called()


def test_interaction_verify_put_unsupported_content_type(monkeypatch, mock_pact, mock_result, fake_interaction):
    monkeypatch.setattr(requests, 'put', Mock())
    monkeypatch.setattr(ResponseVerifier, 'verify', Mock())
    fake_interaction['request']['method'] = 'PUT'
    fake_interaction['request']['headers'] = {'Content-Type': 'spam/ham'}
    i = Interaction(mock_pact('2.0.0'), fake_interaction, mock_result)
    i.verify('http://provider.example/', 'http://provider.example/pact-setup/')
    i.result.fail.assert_called_with("PUT content type spam/ham not implemented in verifier")

    requests.put.assert_not_called()


def test_interaction_verify_patch(monkeypatch, mock_pact, mock_result, fake_interaction):
    monkeypatch.setattr(requests, 'patch', Mock())
    monkeypatch.setattr(requests, 'get', Mock())
    monkeypatch.setattr(ResponseVerifier, 'verify', Mock())
    fake_interaction['request']['method'] = 'PATCH'
    fake_interaction['request']['body'] = 'spam'
    i = Interaction(mock_pact('2.0.0'), fake_interaction, mock_result)
    i.verify('http://provider.example/', 'http://provider.example/pact-setup/')
    requests.get.assert_not_called()
    requests.patch.assert_called_with(
        'http://provider.example/users-service/user/alex',
        json='spam',
        headers={},
    )
    i.response.verify.assert_called()


def test_interaction_verify_patch_unsupported_content_type(monkeypatch, mock_pact, mock_result, fake_interaction):
    monkeypatch.setattr(requests, 'patch', Mock())
    monkeypatch.setattr(ResponseVerifier, 'verify', Mock())
    fake_interaction['request']['method'] = 'PATCH'
    fake_interaction['request']['headers'] = {'Content-Type': 'spam/ham'}
    i = Interaction(mock_pact('2.0.0'), fake_interaction, mock_result)
    i.verify('http://provider.example/', 'http://provider.example/pact-setup/')
    i.result.fail.assert_called_with("PATCH content type spam/ham not implemented in verifier")
    requests.patch.assert_not_called()


@pytest.mark.parametrize('method', ['GET', 'POST', 'DELETE', 'PUT', 'PATCH'])
def test_interaction_sends_headers(monkeypatch, mock_pact, mock_result, fake_interaction, method):
    headers = {'key1': 'value1'}
    requests_method = Mock()
    monkeypatch.setattr(requests, method.lower(), requests_method)
    monkeypatch.setattr(ResponseVerifier, 'verify', Mock())
    fake_interaction['request']['method'] = method
    fake_interaction['request']['body'] = 'body-data'
    fake_interaction['request']['headers'] = headers
    i = Interaction(mock_pact('2.0.0'), fake_interaction, mock_result)
    i.verify('http://provider.example/', 'http://provider.example/pact-setup/')
    request_kwargs = requests_method.call_args[1]
    assert request_kwargs["headers"] == headers


@pytest.mark.parametrize('option, arg', [('providerState', 'state'), ('providerStates', 'states')])
def test_interaction_verify_with_setup(monkeypatch, mock_pact, mock_result, fake_interaction, option, arg):
    monkeypatch.setattr(requests, 'post', Mock(return_value=Mock(status_code=200)))
    monkeypatch.setattr(requests, 'get', Mock())
    monkeypatch.setattr(ResponseVerifier, 'verify', Mock())
    fake_interaction[option] = 'some state'
    i = Interaction(mock_pact('2.0.0'), fake_interaction, mock_result)
    i.verify('http://provider.example/', 'http://provider.example/pact-setup/')

    requests.post.assert_called_with(
        'http://provider.example/pact-setup/',
        json={'provider': "SpamProvider", 'consumer': "SpamConsumer", arg: "some state"}
    )
    i.response.verify.assert_called()


@pytest.mark.parametrize('option', ['providerState', 'providerStates'])
def test_interaction_setup_fails(monkeypatch, mock_pact, mock_result, fake_interaction, option):
    monkeypatch.setattr(requests, 'post', Mock(return_value=Mock(status_code=400, text='fail')))
    monkeypatch.setattr(requests, 'get', Mock())
    monkeypatch.setattr(ResponseVerifier, 'verify', Mock())
    fake_interaction[option] = 'some state'
    i = Interaction(mock_pact('2.0.0'), fake_interaction, mock_result)
    i.verify('http://provider.example/', 'http://provider.example/pact-setup/')
    i.result.fail.assert_called_with("Invalid provider state 'some state'")


def test_interaction_setup_connection_fails(monkeypatch, mock_pact, mock_result, fake_interaction):
    monkeypatch.setattr(requests, 'post', Mock(side_effect=requests.exceptions.ConnectionError('barf')))
    monkeypatch.setattr(requests, 'get', Mock())
    # monkeypatch.setattr(ResponseVerifier, 'verify', Mock())
    fake_interaction['providerState'] = 'some state'
    i = Interaction(mock_pact('2.0.0'), fake_interaction, mock_result)
    i.set_up_state('http://provider.example/pact-setup/', 'state', 'some state')
    i.result.fail.assert_called_once()


def test_response_verifier(fake_interaction, mock_pact):
    fake_interaction['response']['body'] = dict(a='b', c=1)  # note: no rule for matching of c
    fake_interaction['response']['matchingRules'] = {
        '$.body.a': dict(match='type'),
    }
    fake_interaction['response']['headers'] = {'Content-Type': 'json-yeah'}
    r = ResponseVerifier(mock_pact('2.0.0'), fake_interaction['response'], Mock())
    r.verify(Mock(status_code=200, headers={'Content-Type': 'json-yeah'}, json=Mock(return_value=dict(a='b', c='c'))))


def test_build_matching_rules_handles_rule_with_unknown_type_v2(monkeypatch, fake_interaction, mock_pact):
    monkeypatch.setattr(log, 'warning', Mock())
    rules = rule_matchers_v2({
        "$.body": {"match": "SPAM"},
        "$.body[*].*": {"match": "type"}
    })
    print(rules)
    assert 2 == len(rules['body'])
    log.warning.assert_called_once()


def test_build_matching_rules_handles_rule_with_unknown_type_v3(monkeypatch, fake_interaction, mock_pact):
    monkeypatch.setattr(log, 'warning', Mock())
    rules = rule_matchers_v3({
        "body": {
            "$": {"matchers": [{"match": "SPAM"}]},
            "$[*].*": {"matchers": [{"match": "type"}]}
        }
    })
    assert 2 == len(rules['body'])
    log.warning.assert_called_once()


class FakeResponse:
    status = 200
    body = None
    text = 'fake response text'

    def __init__(self, attrs):
        self.__dict__.update(attrs)
        self.status_code = self.status

    def __repr__(self):
        return f'<FakeResponse {self.__dict__}>'

    def json(self):
        return self.body


class FakeRequest:
    method = 'GET'
    body = None

    def __init__(self, attrs):
        self.__dict__.update(attrs)

    def __repr__(self):
        return f'<FakeRequest {self.__dict__}>'

    def json(self):
        return self.body


@pytest.mark.parametrize(
    'filename',
    all_testcases(BASE_DIR / 'pact-spec-version-1.1' / 'testcases' / 'response')
)
def test_version_1_1_response_testcases(filename, mock_pact, mock_result):
    with open(filename) as file:
        case = json.load(file)
        rv = ResponseVerifier(mock_pact('1.1.0'), case['expected'], mock_result)
        rv.verify(FakeResponse(case['actual']))
        success = not bool(rv.result.fail.call_count)
        assert case['match'] == success


@pytest.mark.parametrize(
    'filename',
    all_testcases(BASE_DIR / 'pact-spec-version-1.1' / 'testcases' / 'request')
)
def test_version_1_1_request_testcases(filename, mock_pact, mock_result):
    with open(filename) as file:
        case = json.load(file)
        rv = RequestVerifier(mock_pact('1.1.0'), case['expected'], mock_result)
        rv.verify(FakeRequest(case['actual']))
        success = not bool(rv.result.fail.call_count)
        assert case['match'] == success


@pytest.mark.parametrize(
    'filename',
    all_testcases(BASE_DIR / 'pact-spec-version-2' / 'testcases' / 'response')
)
def test_version_2_response_testcases(filename, mock_pact, mock_result):
    if filename.endswith(' xml.json'):
        # some of the files don't declare the damned content-type!
        raise pytest.skip('XML content type not supported')
    with open(filename) as file:
        case = json.load(file)
        if case['expected'].get('headers', {}).get('Content-Type', "") == 'application/xml':
            raise pytest.skip('XML content type not supported')
        rv = ResponseVerifier(mock_pact('2.0.0'), case['expected'], mock_result)
        rv.verify(FakeResponse(case['actual']))
        success = not bool(rv.result.fail.call_count)
        assert case['match'] == success


@pytest.mark.parametrize(
    'filename',
    all_testcases(BASE_DIR / 'pact-spec-version-2' / 'testcases' / 'request')
)
def test_version_2_request_testcases(filename, mock_pact, mock_result):
    with open(filename) as file:
        case = json.load(file)
        if case['expected'].get('headers', {}).get('Content-Type', "") == 'application/xml':
            raise pytest.skip('XML content type not supported')
        rv = RequestVerifier(mock_pact('2.0.0'), case['expected'], mock_result)
        rv.verify(FakeRequest(case['actual']))
        success = not bool(rv.result.fail.call_count)
        assert case['match'] == success


@pytest.mark.parametrize(
    'filename',
    all_testcases(BASE_DIR / 'pact-spec-version-3' / 'testcases' / 'request')
)
def test_version_3_request_testcases(filename, mock_pact, mock_result):
    with open(filename) as file:
        case = json.load(file)
        if case['expected'].get('headers', {}).get('Content-Type', "") == 'application/xml':
            raise pytest.skip('XML content type not supported')
        rv = RequestVerifier(mock_pact('3.0.0'), case['expected'], mock_result)
        rv.verify(FakeRequest(case['actual']))
        success = not bool(rv.result.fail.call_count)
        assert case['match'] == success


@pytest.mark.parametrize(
    'filename',
    all_testcases(BASE_DIR / 'pact-spec-version-3' / 'testcases' / 'response')
)
def test_version_3_response_testcases(filename, mock_pact, mock_result):
    if filename.endswith(' xml.json'):
        # some of the files don't declare the damned content-type!
        raise pytest.skip('XML content type not supported')
    with open(filename) as file:
        try:
            case = json.load(file)
        except json.JSONDecodeError:
            raise pytest.skip('JSON test case mal-formed')
        if case['expected'].get('headers', {}).get('Content-Type', "") == 'application/xml':
            raise pytest.skip('XML content type not supported')
        rv = ResponseVerifier(mock_pact('3.0.0'), case['expected'], mock_result)
        rv.verify(FakeResponse(case['actual']))
        success = not bool(rv.result.fail.call_count)
        assert case['match'] == success


@pytest.mark.parametrize(
    'filename, verifier, result',
    chain(
        ((t, RequestVerifier, FakeRequest) for t in all_testcases(BASE_DIR / 'testcases-version-3' / 'request')),
        ((t, ResponseVerifier, FakeResponse) for t in all_testcases(BASE_DIR / 'testcases-version-3' / 'response'))
    )
)
def test_local_version_3_testcases(filename, verifier, result, mock_pact, mock_result):
    with open(filename) as file:
        case = json.load(file)
        rv = verifier(mock_pact('3.0.0'), case['expected'], mock_result)
        rv.verify(result(case['actual']))
        success = not bool(rv.result.fail.call_count)
        assert case['match'] == success



import json
import os
import os.path
import urllib.parse

import semver

from ..verifier.verify import RequestVerifier
from ..verifier.result import Result


class Config:
    def __init__(self, consumer_name, provider_name, log_dir, pact_dir, file_write_mode, version):
        self.consumer_name = consumer_name
        self.provider_name = provider_name
        self.log_dir = log_dir
        self.pact_dir = pact_dir
        self.file_write_mode = file_write_mode
        self.version = semver.parse(version)
        self.port = self.allocate_port()
        if file_write_mode == 'overwrite':
            filename = self.pact_filename()
            if os.path.exists(filename):
                os.remove(filename)

    PORT_NUMBER = 8150

    @classmethod
    def allocate_port(cls):
        cls.PORT_NUMBER += 5
        return cls.PORT_NUMBER

    def pact_filename(self):
        return os.path.join(self.pact_dir, f'{self.consumer_name}-{self.provider_name}-pact.json')


class MockPact:
    def __init__(self, config):
        self.provider = config.provider_name
        self.version = config.version


class Request:
    def __init__(self, method, path, query, headers, body):
        self.method = method
        self.path = path
        self.query = query
        self.headers = headers
        self.body = body

    def json(self):
        return self.body


class RecordResult(Result):
    def start(self, interaction):
        super().start(interaction)

    def fail(self, message, path=None):
        self.success = self.FAIL
        self.reason = message
        return not message


class PactRequestHandler:
    def __init__(self, config):
        self.config = config

    def validate_request(self, method):
        url_parts = urllib.parse.urlparse(self.path)

        interaction = self.get_interaction(url_parts.path)

        pact = MockPact(self.config)
        if self.body:
            body = json.loads(self.body)
        else:
            body = ''

        request = Request(method, url_parts.path, url_parts.query, self.headers, body)
        result = RecordResult()
        RequestVerifier(pact, interaction['request'], result).verify(request)
        if not result.success:
            return self.handle_failure(result.reason)
        self.handle_success(interaction)
        self.write_pact(interaction)
        return self.respond_for_interaction(interaction)

    def get_interaction(self, path):
        raise NotImplementedError()

    def handle_success(self, interaction):
        raise NotImplementedError()

    def handle_failure(self, reason):
        raise NotImplementedError()

    def respond_for_interaction(self, reason):
        raise NotImplementedError()

    def write_pact(self, interaction):
        config = self.config
        filename = self.config.pact_filename()
        if os.path.exists(filename):
            with open(filename) as f:
                pact = json.load(f)
            for existing in pact['interactions']:
                if (existing['description'] == interaction['description']
                        and existing['provider_state'] == interaction['provider_state']):
                    # already got one of these...
                    assert existing == interaction, 'Existing "{description}" pact given "{provider_state}" ' \
                        'exists with different request/response'.format(**existing)
                    return
            pact['interactions'].append(interaction)
        else:
            pact = dict(
                consumer=config.consumer_name,
                provider=config.provider_name,
                interactions=[interaction],
                metadata=dict(pactSpecification=dict(version='3.0.0')),
            )

        with open(filename, 'w') as f:
            json.dump(pact, f, indent=2)



from .matchers import get_generated_values, get_matching_rules_v2, get_matching_rules_v3


class Request:
    """Represents an HTTP request and supports Matchers on its properties."""

    def __init__(self, method, path, body=None, headers=None, query=''):
        """
        Create a new instance of Request.

        :param method: The HTTP method that is expected.
        :type method: str
        :param path: The URI path that is expected on this request.
        :type path: str, Matcher
        :param body: The contents of the body of the expected request.
        :type body: str, dict, list
        :param headers: The headers of the expected request.
        :type headers: dict
        :param query: The URI query of the expected request.
        :type query: str or dict
        """
        self.method = method
        self.path = path
        self.body = body
        self.headers = headers
        self.query = query

    def json(self, spec_version):
        """Convert the Request to a JSON Pact."""
        request = {
            'method': self.method,
            'path': get_generated_values(self.path)
        }

        if self.headers:
            request['headers'] = get_generated_values(self.headers)

        if self.body is not None:
            request['body'] = get_generated_values(self.body)

        if self.query:
            request['query'] = get_generated_values(self.query)

        if spec_version == '2.0.0':
            matchingRules = self.generate_v2_matchingRules()
        elif spec_version == '3.0.0':
            matchingRules = self.generate_v3_matchingRules()
        else:
            raise ValueError(f'Invalid Pact specification version={spec_version}')

        if matchingRules:
            request['matchingRules'] = matchingRules

        return request

    def generate_v2_matchingRules(self):
        # TODO check there's generation *and* verification tests for all these
        matchingRules = get_matching_rules_v2(self.path, '$.path')
        matchingRules.update(get_matching_rules_v2(self.headers, '$.headers'))
        matchingRules.update(get_matching_rules_v2(self.body, '$.body'))
        matchingRules.update(get_matching_rules_v2(self.query, '$.query'))
        return matchingRules

    def generate_v3_matchingRules(self):
        # TODO check there's generation *and* verification tests for all these
        matchingRules = get_matching_rules_v3(self.path, 'path')
        matchingRules.update(get_matching_rules_v3(self.headers, 'headers'))

        # body and query rules look different
        body_rules = get_matching_rules_v3(self.body, '$')
        if body_rules:
            matchingRules['body'] = body_rules
        query_rules = get_matching_rules_v3(self.query, '$.query')
        if query_rules:
            matchingRules['query'] = query_rules['$.query']
        return matchingRules



"""Implement verification of pacts as per specification version 2:

https://github.com/pact-foundation/pact-specification/tree/version-2
"""
import json
import logging
import os
import urllib.parse

import semver
from restnavigator import Navigator

from .result import LoggedResult
from .verify import Interaction


log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def pact_id(param):
    return repr(param)


class BrokerPacts:
    def __init__(self, provider_name, pact_broker_url=None, result=LoggedResult()):
        self.provider_name = provider_name
        self.pact_broker_url = pact_broker_url or os.environ.get('PACT_BROKER_URL')
        if not self.pact_broker_url:
            raise ValueError('pact broker URL must be specified')
        self.result = result

    def get_broker_navigator(self):
        # TODO: remove the manipulation of the URL to allow broker URLs with path components at the root
        # (this manipulation is here as an interim measure for older usages which specified a more complex
        # pact broker URL pointing to a different resource)
        url_parts = urllib.parse.urlparse(self.pact_broker_url)
        url = f'{url_parts.scheme}://{url_parts.netloc}/'
        return Navigator.hal(url, default_curie='pb')

    def consumers(self):
        nav = self.get_broker_navigator()
        broker_provider = nav['latest-provider-pacts'](provider=self.provider_name)
        broker_provider.fetch()
        for broker_pact in broker_provider['pacts']:
            pact_contents = broker_pact.fetch()
            yield BrokerPact(pact_contents, self.result, broker_pact)

    def all_interactions(self):
        for pact in self.consumers():
            yield from pact.interactions


class BrokerPact:
    def __init__(self, pact, result, broker_pact=None):
        self.result = result
        self.pact = pact
        self.provider = pact['provider']['name']
        self.consumer = pact['consumer']['name']
        self.metadata = pact['metadata']
        if 'pactSpecification' in self.metadata:
            # the Ruby implementation generates non-compliant metadata, handle that :-(
            self.version = semver.parse(self.metadata['pactSpecification']['version'])
        else:
            self.version = semver.parse(self.metadata['pact-specification']['version'])
        self.interactions = [Interaction(self, interaction, self.result) for interaction in pact['interactions']]
        self.broker_pact = broker_pact

    def __str__(self):
        return f'<Pact consumer={self.consumer} provider={self.provider}>'

    def publish_result(self, version):
        if self.broker_pact is None:
            return
        self.broker_pact['publish-verification-results'].create(dict(success=self.result.success,
                                                                     providerApplicationVersion=version))

    @classmethod
    def load_file(cls, filename, result=LoggedResult()):
        with open(filename) as file:
            return cls(json.load(file), result)



"""Classes and methods to describe contract Consumers."""
import os

from .pact import Pact
from .provider import Provider


USE_MOCKING_SERVER = os.environ.get('PACT_USE_MOCKING_SERVER', 'no') == 'yes'


class Consumer(object):
    """
    A Pact consumer.

    Use this class to describe the service making requests to the provider and
    then use `has_pact_with` to create a contract with a specific service:

    >>> from pactman import Consumer, Provider
    >>> consumer = Consumer('my-web-front-end')
    >>> consumer.has_pact_with(Provider('my-backend-serivce'))
    """

    def __init__(self, name, service_cls=Pact):
        """
        Constructor for the Consumer class.

        :param name: The name of this Consumer. This will be shown in the Pact
            when it is published.
        :type name: str
        :param service_cls: Pact, or a sub-class of it, to use when creating
            the contracts. This is useful when you have a custom URL or port
            for your mock service and want to use the same value on all of
            your contracts.
        :type service_cls: pact.Pact
        """
        self.name = name
        self.service_cls = service_cls

    def has_pact_with(self, provider, host_name='localhost', port=1234,
                      log_dir=None, ssl=False, sslcert=None, sslkey=None,
                      pact_dir=None, version='2.0.0', use_mocking_server=USE_MOCKING_SERVER):
        """
        Create a contract between the `provider` and this consumer.

        If you are running the Pact mock service in a non-default location,
        you can provide the host name and port here:

        >>> from pactman import Consumer, Provider
        >>> consumer = Consumer('my-web-front-end')
        >>> consumer.has_pact_with(
        ...   Provider('my-backend-serivce'),
        ...   host_name='192.168.1.1',
        ...   port=8000)

        :param provider: The provider service for this contract.
        :type provider: pact.Provider
        :param host_name: An optional host name to use when contacting the
            Pact mock service. This will need to be the same host name used by
            your code under test to contact the mock service. It defaults to:
            `localhost`.
        :type host_name: str
        :param port: The TCP port to use when contacting the Pact mock service.
            This will need to tbe the same port used by your code under test
            to contact the mock service. It defaults to: 1234
        :type port: int
        :param log_dir: The directory where logs should be written. Defaults to
            the current directory.
        :type log_dir: str
        :param ssl: Flag to control the use of a self-signed SSL cert to run
            the server over HTTPS , defaults to False.
        :type ssl: bool
        :param sslcert: Path to a custom self-signed SSL cert file, 'ssl'
            option must be set to True to use this option. Defaults to None.
        :type sslcert: str
        :param sslkey: Path to a custom key and self-signed SSL cert key file,
            'ssl' option must be set to True to use this option.
            Defaults to None.
        :type sslkey: str
        :param pact_dir: Directory where the resulting pact files will be
            written. Defaults to the current directory.
        :type pact_dir: str
        :param version: The Pact Specification version to use, defaults to
            '2.0.0'.
        :type version: str
        :param use_mocking_server: If True the mocking will be done using a
            HTTP server rather than patching urllib3 connections. Can be set
            by the environment variable PACT_USE_MOCKING_SERVER which has
            values "no" (default) or "yes".
        :type use_mocking_server: bool
        :return: A Pact object which you can use to define the specific
            interactions your code will have with the provider.
        :rtype: pact.Pact
        """
        if not isinstance(provider, (Provider,)):
            raise ValueError(
                'provider must be an instance of the Provider class.')

        return self.service_cls(
            consumer=self,
            provider=provider,
            host_name=host_name,
            port=port,
            log_dir=log_dir,
            ssl=ssl,
            sslcert=sslcert,
            sslkey=sslkey,
            pact_dir=pact_dir,
            version=version,
            use_mocking_server=use_mocking_server,
        )



"""Implement matching rules as defined in the pact specification version 2:

https://github.com/pact-foundation/pact-specification/tree/version-2
"""
import logging
import re
from collections import OrderedDict, defaultdict

from .paths import format_path


log = logging.getLogger(__name__)


class RuleFailed(Exception):
    def __init__(self, path, message):
        if isinstance(path, list):
            path = format_path(path)
        message = path + ' ' + message
        super().__init__(message)


def split_path(path):
    """Split a JSON path from a pact matchingRule.

    Pact does not support the full JSON path expressions, only ones that match the following rules:

    * All paths start with a dollar ($), representing the root.
    * All path elements are either separated by periods (`.`) or use the JSON path bracket notation (square brackets
      and single quotes around the values: e.g. `['x.y']`), except array indices which use square brackets (`[]`).
      For elements where the value contains white space or non-alphanumeric characters, the JSON path bracket notation
      (`['']`) should be used.
    * The second element of the path is the http type that the matcher is applied to (e.g., $.body or $.header).
    * Path elements represent keys.
    * A star (*) can be used to match all keys of a map or all items of an array (one level only).

    Returns an iterator that has each path element as an item with array indexes converted to integers.
    """
    for elem in re.split(r'[\.\[]', path):
        if elem == '*]':
            yield '*'
        elif elem[0] in "'\"":
            yield elem[1:-2]
        elif elem[-1] == ']':
            yield int(elem[:-1])
        else:
            yield elem


def weight_path(spec_path, element_path):
    """Determine the weighting number for a matchingRule path spec applied to an actual element path from a
    response body.

    Both paths are passed in as lists which represent the paths split per split_path()

    In spec version 2 paths should always start with ['$', 'body'] and contain object element names or array indexes.

    In spec version 3 the path "context" element (like 'body') is moved outside of the path, and the path contains
    just the elements inside the path, always starting at the root '$', so the shortest path is now ['$'].

    Weighting is calculated as:

    * The root node ($) is assigned the value 2.
    * Any path element that does not match is assigned the value 0.
    * Any property name that matches a path element is assigned the value 2.
    * Any array index that matches a path element is assigned the value 2.
    * Any star (*) that matches a property or array index is assigned the value 1.
    * Everything else is assigned the value 0.

    Return the numeric score.
    """
    if len(spec_path) != len(element_path):
        return 0
    score = 1
    for spec, element in zip(spec_path, element_path):
        if spec == element:
            score *= 2
        elif spec == '*':
            score *= 1
        else:
            return 0
    return score


def fold_type(obj):
    if type(obj) == OrderedDict:
        return dict
    return type(obj)


class Matcher:
    """Hold a JSONpath spec and a matchingRule rule and know how to test it.

    The valid rules are:

    `{"match": "regex", "regex": "\\d+"}`
      This executes a regular expression match against the string representation of a values.

    `{"match": "type"}`
       This executes a type based match against the values, that is, they are equal if they are the same type.

    `{"match": "type", "min": 2}`
      This executes a type based match against the values, that is, they are equal if they are the same type.
      In addition, if the values represent a collection, the length of the actual value is compared against the minimum.

    `{"match": "type", "max": 10}`
      This executes a type based match against the values, that is, they are equal if they are the same type.
      In addition, if the values represent a collection, the length of the actual value is compared against the maximum.

    Note that this code
    """
    def __init__(self, path, rule):
        log.debug(f'MatchingRule.__init__ {path!r} {rule!r}')
        self.path = path
        self.rule = rule

    def __repr__(self):
        return f'<{self.__class__.__name__} {self.path} {self.rule}>'

    def __str__(self):
        return f'Rule match by {self.rule} at {self.path}'

    def apply(self, data, spec, path):
        raise NotImplementedError()

    def weight(self, element_path):
        """Given a matching rule path and an element path a blob, determine the weighting for the match.

        Return the weight, or 0 if there is no match.
        """
        return weight_path(list(split_path(self.path)), ['$'] + element_path)

    def check_min(self, data, path):
        if 'min' not in self.rule:
            return
        if type(data) not in (dict, list, str):
            return
        if len(data) < self.rule['min']:
            raise RuleFailed(path, f'size {len(data)!r} is smaller than minimum size {self.rule["min"]}')

    def check_max(self, data, path):
        if 'max' not in self.rule:
            return
        if type(data) not in (dict, list, str):
            return
        if len(data) > self.rule['max']:
            raise RuleFailed(path, f'size {len(data)!r} is larger than maximum size {self.rule["max"]}')

    REGISTRY = {}

    def __init_subclass__(cls, **kwargs):
        if cls not in Matcher.REGISTRY:
            Matcher.REGISTRY[cls.type] = cls
        super().__init_subclass__(**kwargs)

    @classmethod
    def get_matcher(cls, path, rule):
        if 'matchers' in rule:
            # v3 matchingRules always have a matchers array, even if there's a single rule
            return MultipleMatchers(path, **rule)
        if 'regex' in rule:
            # there's a weirdness in the spec here: it promotes use of regex without a match type :(
            type_name = 'regex'
        else:
            type_name = rule.get('match', 'type')
        if type_name not in cls.REGISTRY:
            log.warning(f'invalid match type "{type_name}" in rule at path {path}')
            type_name = 'invalid'
        return cls.REGISTRY[type_name](path, rule)


class InvalidMatcher(Matcher):
    type = 'invalid'

    def apply(self, data, spec, path):
        pass


class MatchType(Matcher):
    type = 'type'

    def apply(self, data, spec, path):
        log.debug(f'match type {data!r} {spec!r} {path!r}')
        if type(spec) in (int, float):
            if type(data) not in (int, float):
                raise RuleFailed(path, f'not correct type ({nice_type(data)} is not {nice_type(spec)})')
        elif fold_type(spec) != fold_type(data):
            raise RuleFailed(path, f'not correct type ({nice_type(data)} is not {nice_type(spec)})')
        self.check_min(data, path)
        self.check_max(data, path)


class MatchRegex(Matcher):
    type = 'regex'

    def apply(self, data, spec, path):
        # we have to cast data to str because Java treats all JSON values as strings and thus is happy to
        # specify a regex matcher for an integer (!!)
        log.debug(f'match regex {data!r} {spec!r} {path!r}: {self.rule["regex"]}')
        if re.fullmatch(self.rule['regex'], str(data)) is None:
            raise RuleFailed(path, f'value {data!r} does not match regex {self.rule["regex"]}')


class MatchInteger(Matcher):
    type = 'integer'

    def apply(self, data, spec, path):
        log.debug(f'match integer {data!r} {spec!r} {path!r}')
        if type(data) != int:
            raise RuleFailed(path, f'not correct type ({nice_type(data)} is not integer)')
        self.check_min(data, path)
        self.check_max(data, path)


class MatchDecimal(Matcher):
    type = 'decimal'

    def apply(self, data, spec, path):
        log.debug(f'match decimal {data!r} {spec!r} {path!r}')
        if type(data) != float:
            raise RuleFailed(path, f'not correct type ({nice_type(data)} is not decimal)')
        self.check_min(data, path)
        self.check_max(data, path)


class MatchNumber(Matcher):
    type = 'number'

    def apply(self, data, spec, path):
        log.debug(f'match number {data!r} {spec!r} {path!r}')
        if type(data) not in (int, float):
            raise RuleFailed(path, f'not correct type ({nice_type(data)} is not number)')
        self.check_min(data, path)
        self.check_max(data, path)


class MatchEquality(Matcher):
    type = 'equality'

    def apply(self, data, spec, path):
        log.debug(f'match equality {data!r} {spec!r} {path!r}')
        if data != self.rule['value']:
            raise RuleFailed(path, f'value {data!r} does not equal expected {self.rule["value"]!r}')


class MatchInclude(Matcher):
    type = 'include'

    def apply(self, data, spec, path):
        log.debug(f'match include {data!r} {spec!r} {path!r}')
        if self.rule['value'] not in data:
            raise RuleFailed(path, f'value {data!r} does not contain expected value {self.rule["value"]!r}')


class MatchNull(Matcher):
    type = 'null'

    def apply(self, data, spec, path):
        log.debug(f'match null {data!r} {spec!r} {path!r}')
        if data is not None:
            raise RuleFailed(path, f'value {data!r} is not null')


class MultipleMatchers(Matcher):
    type = '<multiple>'

    def __init__(self, path, matchers=None, combine='AND'):
        super().__init__(path, matchers)
        self.matchers = [Matcher.get_matcher(path, rule) for rule in matchers]
        self.combine = combine

    def apply(self, data, spec, path):
        log.debug(f'MultipleMatchers.__call__ {data!r} {spec!r} {path!r}')
        for matcher in self.matchers:
            log.debug(f'... matching {matcher}')
            try:
                matcher.apply(data, spec, path)
            except RuleFailed:
                if self.combine == 'AND':
                    raise
            else:
                if self.combine == 'OR':
                    return


def rule_matchers_v2(rules):
    """Get spec v2 rule matchers for the rules sets in a pact's ruleMatchers (passed in as "rules").

    v2 rules are specified in a single dictionary with the jsonpath $.<section>[.additional.jsonpath]:

        "matchingRules": {
            "$.body[0][*].email": {
                "match": "type"
            },
            "$.path": {
                "regex": "/user/\w+/"
            }
        }

    Returns a dict with lists of Matcher subclass instances (e.g. MatchType) for each of path, query, header and body.
    """
    matchers = defaultdict(list)
    for path, spec in rules.items():
        if path == '$.path':
            # "path" rules are a bit different - there's no jsonpath as there's only a single value to compare, so we
            # hard-code the path to '$' which always matches when looking for weighted path matches
            matchers['path'].append(Matcher.get_matcher('$', spec))
        else:
            section = list(split_path(path))[1]
            matchers[section].append(Matcher.get_matcher(path, spec))
    return matchers


def rule_matchers_v3(rules):
    """Get spec v3 rule matchers for the rules sets in a pact's ruleMatchers (passed in as "rules").

    v3 rules are specified in sections with a sub-dict for each of path, query, header and body:

        "matchingRules": {
           "path": {
             "matchers": [
                 { "match": "regex", "regex": "\\w+" }
               ]
           },
           "query": {
             "Q1": {
                 "matchers": [
                   { "match": "regex", "regex": "\\w+" }
                 ]
             }
           },
           "header": {
             "Accept": {
                 "matchers": [
                     { "match" : "regex", "regex" : "\\w+" }
                 ]
             }
           },
           "body": {
             "$.animals": {
                 "matchers": [{"min": 1, "match": "type"}]
             }
           }
        }

    Returns a dict with lists of Matcher subclass instances (e.g. MatchType) for each of path, query, header and body.
    """
    matchers = {}
    if 'path' in rules:
        # "path" rules are a bit different - there's no jsonpath as there's only a single value to compare, so we
        # hard-code the path to '$' which always matches when looking for weighted path matches
        matchers['path'] = [MultipleMatchers('$', **rules['path'])]
    for section in ['query', 'header', 'body']:
        if section in rules:
            matchers[section] = [Matcher.get_matcher(path, rule)
                                 for path, rule in rules[section].items()]
    return matchers


def nice_type(obj):
    """Turn our Python type name into a JSON type name.
    """
    t = fold_type(obj)
    return {
        str: 'string',
        int: 'number',
        float: 'number',
        type(None): 'null',
        list: 'array',
        dict: 'object',
    }.get(t, str(t))



from . import broker_pact


# add the pact broker URL to the pytest output if running verbose
def pytest_report_header(config):
    if config.getoption('verbose') > 0:
        return [f'Loading pacts from {broker_pact.PACT_BROKER_URL}']



from unittest import TestCase

from pactman.mock.matchers import EachLike, Like, Matcher, SomethingLike, \
    Term, from_term, get_generated_values


class MatcherTestCase(TestCase):
    def test_generate(self):
        with self.assertRaises(NotImplementedError):
            Matcher().generate()


class EachLikeTestCase(TestCase):
    def test_default_options(self):
        generate = EachLike(1).generate()

        self.assertEqual(
            generate,
            {'json_class': 'Pact::ArrayLike', 'contents': 1, 'min': 1})

    def test_minimum(self):
        generate = EachLike(1, minimum=2).generate()

        self.assertEqual(
            generate,
            {'json_class': 'Pact::ArrayLike', 'contents': 1, 'min': 2})

    def test_minimum_assertion_error(self):
        with self.assertRaises(AssertionError) as e:
            EachLike(1, minimum=0)

        self.assertEqual(
            str(e.exception), 'Minimum must be greater than or equal to 1')

    def test_nested_matchers(self):
        matcher = EachLike({
            'username': Term('[a-z]+', 'user'),
            'id': SomethingLike(123)})

        generate = matcher.generate()

        self.assertEqual(
            generate,
            {'json_class': 'Pact::ArrayLike',
             'contents': {
                 'username': {
                     'json_class': 'Pact::Term',
                     'data': {
                         'matcher': {
                             'json_class': 'Regexp',
                             's': '[a-z]+',
                             'o': 0},
                         'generate': 'user'}},
                 'id': {
                     'json_class': 'Pact::SomethingLike',
                     'contents': 123}},
             'min': 1})


class LikeTestCase(TestCase):
    def test_is_something_like(self):
        self.assertIs(SomethingLike, Like)


class SomethingLikeTestCase(TestCase):
    def test_valid_types(self):
        types = [None, list(), dict(), 1, 1.0, 'string', u'unicode', Matcher()]

        for t in types:
            SomethingLike(t)

    def test_invalid_types(self):
        with self.assertRaises(AssertionError) as e:
            SomethingLike(set())

        self.assertIn('matcher must be one of ', str(e.exception))

    def test_basic_type(self):
        generate = SomethingLike(123).generate()

        self.assertEqual(
            generate,
            {'json_class': 'Pact::SomethingLike', 'contents': 123})

    def test_complex_type(self):
        generate = SomethingLike({'name': Term('.+', 'admin')}).generate()

        self.assertEqual(
            generate,
            {'json_class': 'Pact::SomethingLike',
             'contents': {'name': {
                 'json_class': 'Pact::Term',
                 'data': {
                     'matcher': {
                         'json_class': 'Regexp',
                         's': '.+',
                         'o': 0
                     },
                     'generate': 'admin'
                 }
             }}})


class TermTestCase(TestCase):
    def test_regex(self):
        generate = Term('[a-zA-Z]', 'abcXYZ').generate()

        self.assertEqual(
            generate,
            {'json_class': 'Pact::Term',
             'data': {
                 'matcher': {
                     'json_class': 'Regexp',
                     's': '[a-zA-Z]',
                     'o': 0},
                 'generate': 'abcXYZ'}})


class FromTermTestCase(TestCase):
    def test_dict(self):
        expected = {'administrator': False, 'id': 123, 'username': 'user'}
        self.assertEqual(from_term(expected), expected)

    def test_none(self):
        self.assertIsNone(from_term(None))

    def test_unicode(self):
        self.assertEqual(from_term(u'testing'), 'testing')

    def test_int(self):
        self.assertEqual(from_term(123), 123)

    def test_float(self):
        self.assertEqual(from_term(3.14), 3.14)

    def test_list(self):
        term = [1, 123, 'sample']
        self.assertEqual(from_term(term), term)

    def test_each_like(self):
        self.assertEqual(
            from_term(EachLike({'a': 1})),
            {'json_class': 'Pact::ArrayLike', 'contents': {'a': 1}, 'min': 1})

    def test_something_like(self):
        self.assertEqual(
            from_term(SomethingLike(123)),
            {'json_class': 'Pact::SomethingLike', 'contents': 123})

    def test_term(self):
        self.assertEqual(
            from_term(Term('[a-f0-9]+', 'abc123')),
            {'json_class': 'Pact::Term',
             'data': {
                 'matcher': {
                     'json_class': 'Regexp',
                     's': '[a-f0-9]+',
                     'o': 0},
                 'generate': 'abc123'}})

    def test_nested(self):
        request = [
            EachLike({
                'username': Term('[a-zA-Z]+', 'firstlast'),
                'id': SomethingLike(123)})]

        self.assertEqual(
            from_term(request),
            [{'contents': {
                'id': {
                    'contents': 123,
                    'json_class': 'Pact::SomethingLike'},
                'username': {
                    'data': {
                        'generate': 'firstlast',
                        'matcher': {
                            'json_class': 'Regexp',
                            'o': 0,
                            's': '[a-zA-Z]+'}},
                        'json_class': 'Pact::Term'}},
                'json_class': 'Pact::ArrayLike',
                'min': 1}])

    def test_unknown_type(self):
        with self.assertRaises(ValueError):
            from_term(set())


class GetGeneratedValuesTestCase(TestCase):
    def test_none(self):
        self.assertIsNone(get_generated_values(None))

    def test_bool(self):
        self.assertFalse(get_generated_values(False))

    def test_unicode(self):
        self.assertEqual(get_generated_values(u'testing'), 'testing')

    def test_int(self):
        self.assertEqual(get_generated_values(123), 123)

    def test_float(self):
        self.assertEqual(get_generated_values(3.14), 3.14)

    def test_list(self):
        term = [1, 123, 'sample']
        self.assertEqual(get_generated_values(term), term)

    def test_dict(self):
        expected = {'administrator': False, 'id': 123, 'username': 'user'}
        self.assertEqual(get_generated_values(expected), expected)

    def test_each_like(self):
        self.assertEqual(
            get_generated_values(EachLike({'a': 1})), [{'a': 1}])

    def test_each_like_minimum(self):
        self.assertEqual(get_generated_values(EachLike({'a': 1}, minimum=5)),
                         [{'a': 1}] * 5)

    def test_something_like(self):
        self.assertEqual(
            get_generated_values(SomethingLike(123)), 123)

    def test_term(self):
        self.assertEqual(
            get_generated_values(Term('[a-f0-9]+', 'abc123')), 'abc123')

    def test_nested(self):
        input = [
            EachLike({
                'username': Term('[a-zA-Z]+', 'firstlast'),
                'id': SomethingLike(123)})]
        self.assertEqual(
            get_generated_values(input),
            [[{'username': 'firstlast', 'id': 123}]])

    def test_unknown_type(self):
        with self.assertRaises(ValueError):
            get_generated_values(set())



"""Classes for defining request and response data that is variable."""


class Matcher(object):
    """Base class for defining complex contract expectations."""

    def generate(self):
        """
        Get the value that the mock service should use for this Matcher.

        :rtype: any
        """
        raise NotImplementedError


class EachLike(Matcher):
    """
    Expect the data to be a list of similar objects.

    Example:

    >>> from pactman import Consumer, Provider
    >>> pact = Consumer('consumer').has_pact_with(Provider('provider'))
    >>> (pact.given('there are three comments')
    ...  .upon_receiving('a request for the most recent 2 comments')
    ...  .with_request('get', '/comment', query={'limit': 2})
    ...  .will_respond_with(200, body={
    ...    'comments': EachLike(
    ...        {'name': SomethingLike('bob'),
    ...         'text': SomethingLike('Hello!')},
    ...        minimum=2)
    ...  }))

    Would expect the response to be a JSON object, with a comments list. In
    that list should be at least 2 items, and each item should be a `dict`
    with the keys `name` and `text`,
    """

    def __init__(self, matcher, minimum=1):
        """
        Create a new EachLike.

        :param matcher: The expected value that each item in a list should
            look like, this can be other matchers.
        :type matcher: None, list, dict, int, float, str, unicode, Matcher
        :param minimum: The minimum number of items expected.
            Must be greater than or equal to 1.
        :type minimum: int
        """
        self.matcher = matcher
        assert minimum >= 1, 'Minimum must be greater than or equal to 1'
        self.minimum = minimum

    def generate(self):
        """
        Generate the value the mock service will return.

        :return: A dict containing the information about the contents of the
            list and the provided minimum number of items for that list.
        :rtype: dict
        """
        return {
            'json_class': 'Pact::ArrayLike',
            'contents': from_term(self.matcher),
            'min': self.minimum}


class Like(Matcher):
    """
    Expect the type of the value to be the same as matcher.

    Example:

    >>> from pactman import Consumer, Provider
    >>> pact = Consumer('consumer').has_pact_with(Provider('provider'))
    >>> (pact
    ...  .given('there is a random number generator')
    ...  .upon_receiving('a request for a random number')
    ...  .with_request('get', '/generate-number')
    ...  .will_respond_with(200, body={
    ...    'number': Like(1111222233334444)
    ...  }))

    Would expect the response body to be a JSON object, containing the key
    `number`, which would contain an integer. When the consumer runs this
    contract, the value `1111222233334444` will be returned by the mock
    service, instead of a randomly generated value.
    """

    def __init__(self, matcher):
        """
        Create a new SomethingLike.

        :param matcher: The object that should be expected. The mock service
            will return this value. When verified against the provider, the
            type of this value will be asserted, while the value will be
            ignored.
        :type matcher: None, list, dict, int, float, str, Matcher
        """
        valid_types = (
            type(None), list, dict, int, float, str, Matcher)

        assert isinstance(matcher, valid_types), (
            "matcher must be one of '{}', got '{}'".format(
                valid_types, type(matcher)))

        self.matcher = matcher

    def generate(self):
        """
        Return the value that should be used in the request/response.

        :return: A dict containing the information about what the contents of
            the response should be.
        :rtype: dict
        """
        return {
            'json_class': 'Pact::SomethingLike',
            'contents': from_term(self.matcher)}


# Remove SomethingLike in major version 1.0.0
SomethingLike = Like


class Term(Matcher):
    """
    Expect the response to match a specified regular expression.

    Example:

    >>> from pactman import Consumer, Provider
    >>> pact = Consumer('consumer').has_pact_with(Provider('provider'))
    >>> (pact.given('the current user is logged in as `tester`')
    ...  .upon_receiving('a request for the user profile')
    ...  .with_request('get', '/profile')
    ...  .will_respond_with(200, body={
    ...    'name': 'tester',
    ...    'theme': Term('light|dark|legacy', 'dark')
    ...  }))

    Would expect the response body to be a JSON object, containing the key
    `name`, which will contain the value `tester`, and `theme` which must be
    one of the values: light, dark, or legacy. When the consumer runs this
    contract, the value `dark` will be returned by the mock service.
    """

    def __init__(self, matcher, generate):
        """
        Create a new Term.

        :param matcher: A regular expression to find.
        :type matcher: basestring
        :param generate: A value to be returned by the mock service when
            generating the response to the consumer.
        :type generate: basestring
        """
        self.matcher = matcher
        self._generate = generate

    def generate(self):
        """
        Return the value that should be used in the request/response.

        :return: A dict containing the information about what the contents of
            the response should be, and what should match for the requests.
        :rtype: dict
        """
        return {
            'json_class': 'Pact::Term',
            'data': {
                'generate': self._generate,
                'matcher': {
                    'json_class': 'Regexp',
                    'o': 0,
                    's': self.matcher}}}


def from_term(term):
    """
    Parse the provided term into the JSON for the mock service.

    :param term: The term to be parsed.
    :type term: None, list, dict, int, float, str, unicode, Matcher
    :return: The JSON representation for this term.
    :rtype: dict, list, str
    """
    if term is None:
        return term
    elif isinstance(term, (str, int, float)):
        return term
    elif isinstance(term, dict):
        return {k: from_term(v) for k, v in term.items()}
    elif isinstance(term, list):
        return [from_term(t) for i, t in enumerate(term)]
    elif issubclass(term.__class__, (Matcher,)):
        return term.generate()
    else:
        raise ValueError('Unknown type: %s' % type(term))


def get_generated_values(input):
    """
    Resolve (nested) Matchers to their generated values for assertion.

    :param input: The input to be resolved to its generated values.
    :type input: None, list, dict, int, float, bool, str, unicode, Matcher
    :return: The input resolved to its generated value(s)
    :rtype: None, list, dict, int, float, bool, str, unicode, Matcher
    """
    if input is None:
        return input
    if isinstance(input, (str, int, float, bool)):
        return input
    if isinstance(input, dict):
        return {k: get_generated_values(v) for k, v in input.items()}
    if isinstance(input, list):
        return [get_generated_values(t) for i, t in enumerate(input)]
    elif isinstance(input, Like):
        return get_generated_values(input.matcher)
    elif isinstance(input, EachLike):
        return [get_generated_values(input.matcher)] * input.minimum
    elif isinstance(input, Term):
        return input.generate()['data']['generate']
    else:
        raise ValueError('Unknown type: %s' % type(input))


def get_matching_rules_v2(input, path):
    if input is None or isinstance(input, (str, int, float, bool)):
        return {}
    if isinstance(input, dict):
        rules = {}
        for k, v in input.items():
            sub_path = path + '.' + k
            rules.update(get_matching_rules_v2(v, sub_path))
        return rules
    if isinstance(input, list):
        rules = {}
        for i, v in enumerate(input):
            sub_path = f'{path}[*]'
            rules.update(get_matching_rules_v2(v, sub_path))
        return rules
    elif isinstance(input, Like):
        rules = {path: {'match': 'type'}}
        rules.update(get_matching_rules_v2(input.matcher, path))
        return rules
    elif isinstance(input, EachLike):
        rules = {path: {'match': 'type', 'min': input.minimum}}
        rules.update(get_matching_rules_v2(input.matcher, path))
        return rules
    elif isinstance(input, Term):
        return {path: {'regex': input.matcher}}
    else:
        raise ValueError('Unknown type: %s' % type(input))


def get_matching_rules_v3(input, path):
    if input is None or isinstance(input, (str, int, float, bool)):
        return {}
    if isinstance(input, dict):
        rules = {}
        for k, v in input.items():
            sub_path = path + '.' + k
            rules.update(get_matching_rules_v3(v, sub_path))
        return rules
    if isinstance(input, list):
        rules = {}
        for i, v in enumerate(input):
            sub_path = f'{path}[*]'
            rules.update(get_matching_rules_v3(v, sub_path))
        return rules

    if not path.startswith('$.body'):
        raise ValueError(f'Pact specification v2 does not allow non-body {input}')

    if isinstance(input, Like):
        rules = {path: {'matchers': [{'match': 'type'}]}}
        rules.update(get_matching_rules_v3(input.matcher, path))
        return rules
    elif isinstance(input, EachLike):
        rules = {path: {'matchers': [{'match': 'type', 'min': input.minimum}]}}
        rules.update(get_matching_rules_v3(input.matcher, path))
        return rules
    elif isinstance(input, Term):
        return {path: {'matchers': [{'match': 'regex', 'regex': input.matcher}]}}
    else:
        raise ValueError('Unknown type: %s' % type(input))



"""Python methods for interactive with a Pact Mock Service."""
from .mock.consumer import Consumer
from .mock.matchers import EachLike, Like, SomethingLike, Term
from .mock.pact import Pact
from .mock.provider import Provider

__all__ = ('Consumer', 'EachLike', 'Like', 'Pact', 'Provider', 'SomethingLike',
           'Term')



from .matchers import get_generated_values, get_matching_rules_v2, get_matching_rules_v3


class Response:
    """Represents an HTTP response and supports Matchers on its properties."""

    def __init__(self, status, headers=None, body=None):
        """
        Create a new Response.

        :param status: The expected HTTP status of the response.
        :type status: int
        :param headers: The expected headers of the response.
        :type headers: dict
        :param body: The expected body of the response.
        :type body: str, dict, or list
        """
        self.status = status
        self.body = body
        self.headers = headers

    def json(self, spec_version):
        """Convert the Response to a JSON Pact."""
        response = {'status': self.status}
        if self.body is not None:
            response['body'] = get_generated_values(self.body)

        if self.headers:
            response['headers'] = get_generated_values(self.headers)

        if spec_version == '2.0.0':
            matchingRules = self.generate_v2_matchingRules()
        elif spec_version == '3.0.0':
            matchingRules = self.generate_v3_matchingRules()
        else:
            raise ValueError(f'Invalid Pact specification version={spec_version}')

        if matchingRules:
            response['matchingRules'] = matchingRules

        return response

    def generate_v2_matchingRules(self):
        # TODO check there's generation *and* verification tests for all these
        matchingRules = get_matching_rules_v2(self.headers, '$.headers')
        matchingRules.update(get_matching_rules_v2(self.body, '$.body'))
        return matchingRules

    def generate_v3_matchingRules(self):
        # TODO check there's generation *and* verification tests for all these
        matchingRules = get_matching_rules_v3(self.headers, 'headers')
        body_rules = get_matching_rules_v3(self.body, '$')
        if body_rules:
            matchingRules['body'] = body_rules
        return matchingRules
